package example_counting_happywords_set;

import java.util.*;
import java.io.*;

public class CountKeywords {
	// Array of all happy words
	static String[] happyWordsAry = {"content", "cheerful", "cheery",
	    "merry", "joyful", "joy", "jovial", "jolly", "joking", "joke", "jocular", "gleeful",
	    "glee", "carefree", "untroubled", "delighted", "delight", "smiling", "smile", 
	    "delighted", "delight", "elated", "elation", 
	    "glad", "joyous", "jubilant", "lively", "pleased", "thrilled", "happy",
	    "upbeat", "blessed", "blest", "blissful", "chipper", "chirpy", "content",
	    "convivial", "gay", "gratified", "laughing", "laugh", "mirthful", "peppy",
	    "playful"};
	
	static Set<String> happyWords = new HashSet<>(Arrays.asList(happyWordsAry));
    static List<String> happyWords2 = new ArrayList<>(Arrays.asList(happyWordsAry));

	public static void main(String[] args) throws Exception {
    	String filename = "src/example_counting_happywords_set/Catcher_in_the_Rye.txt";
	    File file = new File(filename);

	    displayFirstWords(file); // for debug, show first 200 words in file
	    displayKeywordCount(file);
	    displayKeywordsUsedInOrder(file);
	    displayKeywordsUsedSorted(file);

	    System.out.println("\nSupplemental, not used in text:");
	    displayKeywordCount2(file);
	    displayKeywordsUsedSortedVer2(file);
	    displayKeywordsUsedSortedVer3(file);
	}

	//--------------------------------------------------------------
	// Display methods, Utility methods in section below.
	//--------------------------------------------------------------

	public static void displayFirstWords(File file) throws Exception {
	    System.out.println("\n\nFirst 200 words:");
	    Set<String> words = getWords(file, 200);
	    int i=0;
    	for(String word : words) {
    		System.out.print(word + ", ");
    		if((++i)%15==0) {
    			System.out.println();
    		}
    	}
	}
	
	public static void displayKeywordCount(File file) throws Exception {
	  	long begTime = System.currentTimeMillis();
	  	int count = countKeywords(file, happyWords);
    	long endTime = System.currentTimeMillis();

    	System.out.println("\n\nThe number of keywords in " + file.getName() + " is " + count);
    	
    	double totTime = (endTime-begTime)/1000.0;
	    String msg = String.format("Time=%.3f sec", totTime);
	    System.out.println(msg);
	}
	
	public static void displayKeywordsUsedInOrder(File file) throws Exception {
    	System.out.println("\nHappy words in file, in order of occurrence:");

    	long begTime = System.currentTimeMillis();
    	Set<String> happyWordsUsedInOrder = getWordsUsedInOrder(file, happyWords);
    	long endTime = System.currentTimeMillis();

    	for(String word : happyWordsUsedInOrder) System.out.print(word + ", ");
    	double totTime = (endTime-begTime)/1000.0;
	    String msg = String.format("\nTime=%.3f sec", totTime);
	    System.out.println(msg);
	}

	public static void displayKeywordsUsedSorted(File file) throws Exception {
	   	System.out.println("\nHappy words in file, sorted:");

    	long begTime = System.currentTimeMillis();
    	Set<String> happyWordsUsedSorted = getWordsUsedSorted(file, happyWords);
    	long endTime = System.currentTimeMillis();

    	for(String word : happyWordsUsedSorted) System.out.print(word + ", ");

    	double totTime = (endTime-begTime)/1000.0;
	    String msg = String.format("\nTime=%.3f sec", totTime);
	    System.out.println(msg);
	}

	//--------------------------------------------------------------
	// Utility methods
	//--------------------------------------------------------------
	public static Set<String> getWords(File file, int num) throws Exception {
	    int count = 0;
	    Scanner input = new Scanner(file);
	    Set<String> words = new LinkedHashSet<>();
	    while (input.hasNext()) {
	    	String word = input.next();
	    	word = getCleanWord(word);
	    	words.add(word);
	    	count++;
	    	if(count > num) {
	    		break;
	    	}
	    }
	    input.close();
	    return words;
	}
	
	public static int countKeywords(File file, Set<String> happyWords) throws Exception {
	    int count = 0;
	    Scanner input = new Scanner(file);
	
	    while (input.hasNext()) {
	    	String word = input.next();
	    	word = getCleanWord(word);
	    	if (happyWords.contains(word)) {
	    		count++;
	    	}
	    }
	    input.close();
	    return count;
	}
	
	public static String getCleanWord(String word) {
		// Misses (using "*" as delimiter): *your--*, *you--a*, *--Hey*
		// In other words, leading, embedded, or trailing "--".
		// Also misses: *'The*, *'no'*, *'Digression!'*
		// In other words, leading and trailing "'"
		String punctuation = ".,;:\"?!";
		String cleanWord = word.trim();
		cleanWord.toLowerCase();
		boolean isFinished = false;
		// Remove trailing punctuation. Will solve (among other cases): dog.", e.g. end of a quoted sentence.
		while(!isFinished) {
			if(cleanWord.length()>0) {
				String lastChar = String.valueOf(cleanWord.charAt(cleanWord.length()-1));
				if(punctuation.contains(lastChar)) {
					cleanWord = cleanWord.substring(0,cleanWord.length()-1);
				}
				else {
					isFinished = true;
				}
			}
			else {
				isFinished = true;
			}
		}
		// Remove leading punctuation.
		isFinished = false;
		while(!isFinished) {
			if(cleanWord.length()>0) {
				String firstChar = String.valueOf(cleanWord.charAt(0));
				if(punctuation.contains(firstChar)) {
					cleanWord = cleanWord.substring(1);
				}
				else {
					isFinished = true;
				}
			}
			else {
				isFinished = true;
			}
		}
		return cleanWord;
	}
	
	public static Set<String> getWordsUsedSorted(File file, Set<String> happyWords) throws FileNotFoundException {
		Set<String> happyWordsUsed = new TreeSet<>();
	    Scanner input = new Scanner(file);
	    while (input.hasNext()) {
		    	String word = input.next();
		    	word = getCleanWord(word);
		    	if (happyWords.contains(word))
		    		if(!happyWordsUsed.contains(word))
			    		happyWordsUsed.add(word);
	    }
	    input.close();
	
		return happyWordsUsed;
	}

	public static Set<String> getWordsUsedInOrder(File file, Set<String> happyWords) throws FileNotFoundException {
		Set<String> happyWordsUsed = new LinkedHashSet<>();
	    Scanner input = new Scanner(file);
	
	    while (input.hasNext()) {
	    	String word = input.next();
	    	word = getCleanWord(word);
	    	if (happyWords.contains(word)) {
	    		if(!happyWordsUsed.contains(word)) { // not necessary but makes more understandable
	    			happyWordsUsed.add(word);
	    		}
	    	}
	    }
	    input.close();
	
	    return happyWordsUsed;
	}

	//--------------------------------------------------------------
	// Supplemental Display methods, not used in text
	//--------------------------------------------------------------
	public static void displayKeywordCount2(File file) throws Exception {
	  	long begTime = System.currentTimeMillis();
	  	int count = countKeywords2(file, happyWords);
    	long endTime = System.currentTimeMillis();

    	System.out.println("\nThe number of keywords in " + file.getName() + " is " + count);
    	double totTime = (endTime-begTime)/1000.0;
	    String msg = String.format("Time=%.3f sec", totTime);
	    System.out.println(msg);
	}

	public static void displayKeywordsUsedSortedVer2(File file) throws Exception {
    	System.out.println("\nHappy words in file, sorted, ver 2:");

    	long begTime = System.currentTimeMillis();
    	Set<String> happyWordsUsedSorted2 = getWordsUsedSorted2(file, happyWords);
    	long endTime = System.currentTimeMillis();

    	for(String word : happyWordsUsedSorted2) System.out.print(word + ", ");

    	double totTime = (endTime-begTime)/1000.0;
	    String msg = String.format("\nTime=%.3f sec", totTime);
	    System.out.println(msg);
	}

	public static void displayKeywordsUsedSortedVer3(File file) throws Exception {
    	System.out.println("\nHappy words in file, sorted, ver 3:");

    	long begTime = System.currentTimeMillis();
    	Set<String> happyWordsUsedSorted3 = getWordsUsedSorted3(file, happyWords);
    	long endTime = System.currentTimeMillis();

    	for(String word : happyWordsUsedSorted3) System.out.print(word + ", ");

    	double totTime = (endTime-begTime)/1000.0;
	    String msg = String.format("\nTime=%.3f sec", totTime);
	    System.out.println(msg);
	}

	//--------------------------------------------------------------
	// Supplemental Utility methods, not used in text
	//--------------------------------------------------------------
	public static int countKeywords2(File file, Set<String> happyWords) throws Exception {
	    int count = 0;
	    Scanner input = new Scanner(file);
	
	    while (input.hasNext()) {
	    	String word = input.next();
	    	word = getCleanWord(word);
	    	if (happyWords2.contains(word))
	    		count++;
	    }
	    input.close();
	    return count;
	}
	
	public static Set<String> getWordsUsedSorted2(File file, Set<String> happyWords) throws FileNotFoundException {
	    Scanner input = new Scanner(file);
	    input.useDelimiter("\\Z");
	    String text = input.next();
	    input.close();
	    String[] words = text.split("[\\s.,;:\"?!]+");
	    Set<String> uWords = new TreeSet<>(Arrays.asList(words));
	    uWords.retainAll(happyWords);
		return uWords;
	}
	
	public static Set<String> getWordsUsedSorted3(File file, Set<String> happyWords) throws FileNotFoundException {
		Set<String> uWords = new TreeSet<>(happyWords);
	    Scanner input = new Scanner(file);
	    input.useDelimiter("\\Z");
	    String text = input.next();
	    input.close();
	    String[] words = text.split("[\\s.,;:\"?!]+");
	    List<String> listWords = new LinkedList<>(Arrays.asList(words));
	    uWords.retainAll(listWords);
		return uWords;
	}
}
