package example_iterator_cities;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;

public class IteratorExample {
	public static void main(String[] args) {
		traversal();
		removeWithIterator_Correct();
		removeWithBackwardsLoop_Correct();
		removeWithWhileLoop_Correct();
		removeWithForwardLoop_Correct();
		removeWithForwardLoop_Incorrect();
		// This method will generate a runtime error
		//removeWithForeachLoop_Incorrect();
		removeAndAddWithIterator_Correct();
		removeAndAddWithIterator_Incorrect();
		removeWithOddSyntaxIterator_Correct();
		removeWithRemoveIf_Correct();
		removeWithStreamFilter_Correct();
		oddIndexedLoops();
	}
	
	public static void traversal() {
		System.out.println("\n-->Traveral with iterator, foreach, indexed for");
		ArrayList<String> cities = buildCitiesList();

		// Iterator
		Iterator<String> iter = cities.iterator();
		while(iter.hasNext()) {
		    String city = iter.next();
		    System.out.println(city);
		}
		System.out.println();

		// foreach loop
		for(String city : cities) {
		    System.out.println(city);
		}
		System.out.println();


		// indexed for loop
		for(int i=0; i<cities.size(); i++) {
		    System.out.println(cities.get(i));
		}
		System.out.println();
	}
	
	public static void removeWithIterator_Correct() {
		System.out.println("\n-->Remove with iterator");
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");

		// iterator()
		Iterator<String> iter = cities.iterator();
		while( iter.hasNext() ) {
			String city = iter.next();
			if(city.equals("New York")) {
				iter.remove();
			}
		}
		printCities(cities, "After removing New York");
	}
	
	public static void removeWithBackwardsLoop_Correct() {
		System.out.println("\n-->Remove with loop, correct");
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");
		// loop approach, must go backwards through loop
		for(int i=cities.size()-1; i>=0; i--) {
			if(cities.get(i).equals("New York")) {
				cities.remove(i);
			}
		}
		printCities(cities, "After removing New York");
	}
	
	public static void removeWithForwardLoop_Incorrect() {
		System.out.println("\n-->Remove with loop, incorrect");
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");
		// loop approach, will skip the second New York
		for(int i=0; i<cities.size(); i++) {
			if(cities.get(i).equals("New York")) {
				cities.remove(i);
			}
		}
		printCities(cities, "After removing New York");
	}

	public static void removeWithForwardLoop_Correct() {
		System.out.println("\n-->Remove with forward loop, correct");
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");
		for(int i=0; i<cities.size(); i++) {
			if(cities.get(i).equals("New York")) {
				cities.remove(i--); // decrement i after removing, so doesn't skip
			}
		}
		printCities(cities, "After removing New York");
	}

	public static void removeWithWhileLoop_Correct() {
		System.out.println("\n-->Remove with while loop, correct");
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");
		int i=0;
		while(i < cities.size()) {
			if(cities.get(i).equals("New York")) {
				cities.remove(i);
			}
			else {
				i++;
			}
		}
		printCities(cities, "After removing New York");
	}

	// This will generate runtime error if executed
	public static void removeWithForeachLoop_Incorrect() {
		System.out.println("\n-->Remove with foreach loop, incorrect");
		
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");
		// foreach loop approach
		for(String city : cities) {
			if(city.equals("New York")) {
				cities.remove(city);
			}
		}
		printCities(cities, "After removing New York");
	}

	/*
	 * Show how to remove from one list, and add to another.
	 */
	public static void removeAndAddWithIterator_Correct() {
		System.out.println("\n-->Remove and add with iterator, correct");
		
		ArrayList<String> cities = buildCitiesList();
		ArrayList<String> citiesNY = new ArrayList<String>();

		printCities(cities, "Before removing New York");
		// iterator()
		Iterator<String> iter = cities.iterator();
		while( iter.hasNext() ) {
			String city = iter.next();
			if(city.equals("New York")) {
				iter.remove();
				citiesNY.add(city);
			}
		}
		printCities(cities, "cities: After removing New York");
		printCities(citiesNY, "citiesNY:");
	}

	/*
	 * Show how to remove from one list, and add to another, incorrectly!.
	 * Will skip the second New Yorks b/c iter.next is called twice within
	 * the loop. 
	 */
	public static void removeAndAddWithIterator_Incorrect() {
		System.out.println("\n-->Remove and add with iterator, incorrect");
		
		ArrayList<String> cities = buildCitiesList();
		ArrayList<String> citiesNY = new ArrayList<>();
		
		printCities(cities, "Before removing New York");
		// iterator()
		Iterator<String> iter = cities.iterator();
		while( iter.hasNext() ) {
			if(iter.next().equals("New York")) {
				iter.remove();
				citiesNY.add(iter.next());
			}
		}
		printCities(cities, "cities: After removing New York");
		printCities(citiesNY, "citiesNY:");
	}
	
	public static void removeWithRemoveIf_Correct() {
		System.out.println("\n-->Remove with Stream filter, correct");
		ArrayList<String> cities = buildCitiesList();
		printCities(cities, "Before removing New York");
		
		cities.removeIf(x -> x.equals("New York"));
		
		printCities(cities, "After removing New York");
	}

	public static void removeWithStreamFilter_Correct() {
		System.out.println("\n-->Remove with Stream filter, correct");
		ArrayList<String> cities = buildCitiesList();
		printCities(cities, "Before removing New York");
	
		List<String> duplicates = cities
		          .stream()
		          .filter(x -> x.equals("New York"))
		          .collect(Collectors.toList());
		
		cities.removeAll(duplicates);

		printCities(duplicates, "After removing New York");
		printCities(cities, "After removing New York");
	}

	/*
	 * Shows several loops with constructs we did not consider
	 */
	public static void removeWithOddSyntaxIterator_Correct() {
		System.out.println("\n-->removeWithOddSyntaxIterator");
		ArrayList<String> cities = buildCitiesList();

		printCities(cities, "Before removing New York");
		// Uses iterator in the for statement itself.
		for(Iterator<String> iter = cities.iterator(); iter.hasNext(); ) {
		     if(iter.next().equals("New York")) {
		    	 iter.remove();
		     }
		}
		printCities(cities, "After removing New York");
		
	}

	public static void oddIndexedLoops() {
		System.out.println("\n-->oddIndexedLoops");
		ArrayList<String> cities = buildCitiesList();
		
		// Increment is in the loop. Not a good approach
		for(int i=0; i<cities.size(); ) {
			System.out.println(cities.get(i++));
		}
		
		// Illustrates how you can use multiple variables in a for loop.
		// Doesn't relate to anything in this chapter
		for(int i=0, j=0; i<4; i++, j+=2) {
			System.out.println(i + ", " + j);
		}
	}

	private static ArrayList<String> buildCitiesList() {
		ArrayList<String> cities = new ArrayList<String>();
		cities.add("Dallas");
		cities.add("New York");
		cities.add("New York");
		cities.add("San Franciso");
		cities.add("Madison");
		return cities;
	}
	
	private static void printCities(List<String> cities, String msg) {
		System.out.println("\n" + msg);
		System.out.println(cities);

	}

}
