package example_iterator_players;

import java.util.ArrayList;
import java.util.Iterator;

public class IteratorExample {
	public static void main(String[] args) {
		traversal();
		removeWithIterator_Correct();
		removeWithBackwardsLoop_Correct();
		removeWithWhileLoop_Correct();
		removeWithForwardLoop_Correct();
		removeWithForwardLoop_Incorrect();
		// This method will generate a runtime error
		//removeWithForeachLoop_Incorrect();
		removeAndAddWithIterator_Correct();
		removeAndAddWithIterator_Incorrect();
		removeWithOddSyntaxIterator_Correct();
		oddIndexedLoops();
	}
	
	public static void traversal() {
		System.out.println("\n-->Traveral with iterator, foreach, indexed for");
		ArrayList<Player> players = buildPlayersList();

		// Iterator
		Iterator<Player> iter = players.iterator();
		while(iter.hasNext()) {
			Player player = iter.next();
		    System.out.println(player);
		}
		System.out.println();

		// foreach loop
		for(Player player : players) {
		    System.out.println(player);
		}
		System.out.println();
		
		
		// indexed for loop
		for(int i=0; i<players.size(); i++) {
		    System.out.println(players.get(i));
		}
		System.out.println();
	}
	
	public static void removeWithIterator_Correct() {
		System.out.println("\n-->Remove with iterator");
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		// iterator()
		Iterator<Player> iter = players.iterator();
		while(iter.hasNext() ) {
			Player player = iter.next();
			if(player.getScore() < 20) {
				iter.remove();
			}
		}
		printPlayers(players, "After removing players with score<20");
	}
	
	public static void removeWithBackwardsLoop_Correct() {
		System.out.println("\n-->Remove with loop, correct");
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		// loop approach, must go backwards through loop
		for(int i=players.size()-1; i>=0; i--) {
			if(players.get(i).equals("New York")) {
				players.remove(i);
			}
		}
		printPlayers(players, "After removing players with score<20");

	}
	
	public static void removeWithForwardLoop_Incorrect() {
		System.out.println("\n-->Remove with loop, incorrect");
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		// loop approach, will skip the second New York
		for(int i=0; i<players.size(); i++) {
			if(players.get(i).equals("New York")) {
				players.remove(i);
			}
		}
		printPlayers(players, "After removing players with score<20");

	}

	public static void removeWithForwardLoop_Correct() {
		System.out.println("\n-->Remove with forward loop, correct");
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		for(int i=0; i<players.size(); i++) {
			if(players.get(i).equals("New York")) {
				players.remove(i--); // decrement i after removing, so doesn't skip
			}
		}
		printPlayers(players, "After removing players with score<20");

	}

	public static void removeWithWhileLoop_Correct() {
		System.out.println("\n-->Remove with while loop, correct");
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		int i=0;
		while(i < players.size()) {
			if(players.get(i).equals("New York")) {
				players.remove(i);
			}
			else {
				i++;
			}
		}
		printPlayers(players, "After removing players with score<20");

	}

	// This will generate runtime error if executed
	public static void removeWithForeachLoop_Incorrect() {
		System.out.println("\n-->Remove with foreach loop, incorrect");
		
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		// foreach loop approach
		for(Player player : players) {
			if(player.equals("New York")) {
				players.remove(player);
			}
		}
		printPlayers(players, "After removing players with score<20");

	}

	/*
	 * Show how to remove from one list, and add to another.
	 */
	public static void removeAndAddWithIterator_Correct() {
		System.out.println("\n-->Remove and add with iterator, correct");
		
		ArrayList<Player> players = buildPlayersList();
		ArrayList<Player> lowScorePlayers = new ArrayList<>();
		
		printPlayers(players, "Original List");

		// iterator()
		Iterator<Player> iter = players.iterator();
		while( iter.hasNext() ) {
			Player player = iter.next();
			if(player.getScore() < 20) {
				iter.remove();
				lowScorePlayers.add(player);
			}
		}
		printPlayers(players, "players: After removing players with score<20");
		printPlayers(lowScorePlayers, "lowScorePlayers:");
	}

	/*
	 * Show how to remove from one list, and add to another, incorrectly!.
	 * Will skip the second New Yorks b/c iter.next is called twice within
	 * the loop. 
	 */
	public static void removeAndAddWithIterator_Incorrect() {
		System.out.println("\n-->Remove and add with iterator, incorrect");
		
		ArrayList<Player> players = buildPlayersList();
		ArrayList<Player> lowScorePlayers = new ArrayList<>();
		
		printPlayers(players, "Original List");

		// iterator()
		Iterator<Player> iter = players.iterator();
		while( iter.hasNext() ) {
			if(iter.next().getScore() < 20) {
				iter.remove();
				lowScorePlayers.add(iter.next());
			}
		}
		printPlayers(players, "players: After removing players with score<20");
		printPlayers(lowScorePlayers, "lowScorePlayers:");
	}

	/*
	 * Shows several loops with constructs we did not consider
	 */
	public static void removeWithOddSyntaxIterator_Correct() {
		System.out.println("\n-->removeWithOddSyntaxIterator");
		ArrayList<Player> players = buildPlayersList();

		printPlayers(players, "Original List");

		// Uses iterator in the for statement itself.
		for(Iterator<Player> iter = players.iterator(); iter.hasNext(); ) {
		     if(iter.next().equals("New York")) {
		    	 iter.remove();
		     }
		}
		printPlayers(players, "After removing players with score<20");
	}

	public static void oddIndexedLoops() {
		System.out.println("\n-->oddIndexedLoops");
		ArrayList<Player> cities = buildPlayersList();
		
		// Increment is in the loop. Not a good approach
		for(int i=0; i<cities.size(); ) {
			System.out.println(cities.get(i++));
		}
		
		// Illustrates how you can use multiple variables in a for loop.
		// Doesn't relate to anything in this chapter
		for(int i=0, j=0; i<4; i++, j+=2) {
			System.out.println(i + ", " + j);
		}
	}

	private static ArrayList<Player> buildPlayersList() {
		
		ArrayList<Player> players = new ArrayList<>();
		players.add(new Player("Pam", 24));
		players.add(new Player("Len", 19));
		players.add(new Player("Malia", 37));
		players.add(new Player("Bob", 13));
		players.add(new Player("Rea", 46));
		return players;
	}
	
	private static void printPlayers(ArrayList<Player> cities, String msg) {
		System.out.println("\n" + msg);
		System.out.println(cities);

	}
}
